import argparse
import sys

from pathlib import Path
from typing import List, Dict

from yaml_extender.resolver import reference_resolver
from yaml_extender.xyml_file import XYmlFile
from yaml_extender.logger import get_logger

LOGGER = get_logger()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input", help="Input yaml file to be parsed", type=Path)
    parser.add_argument("output", help="Output file to save to", type=Path)
    parser.add_argument("-i", "--include", help="Include paths", type=Path, nargs="*")
    args, unknown_args = parser.parse_known_args()

    if not args.input.is_file:
        raise FileNotFoundError(f"Path {args.input} is no valid file.")
    additional_args = parse_unknown_args(unknown_args)
    LOGGER.info("Additional parameters:\n" + "\n".join([f"{k}: {v}" for k, v in additional_args.items()]))
    xyml_file = XYmlFile(args.input, additional_args)
    output_dir: Path = args.output.parent
    output_dir.mkdir(exist_ok=True, parents=True)
    xyml_file.save(args.output)


def parse_unknown_args(args: List) -> Dict:
    arg_dict = dict(zip(args[:-1:2], args[1::2]))
    ret_val = {}
    # Parse numbers from strings
    for k, v in arg_dict.items():
        ret_val[k.strip("-")] = reference_resolver.number_parse(v)
    return ret_val


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

import abc
from pathlib import Path
from typing import Any


class Resolver(abc.ABC):

    def __init__(self, fail_on_resolve: bool = True):
        """
        Parameters
            fail_on_resolve: Flag if the parsing should be aborted if a value fails to resolve
        """
        self.fail_on_resolve: bool = fail_on_resolve
        super().__init__()

    def resolve(self, content: Any, config: dict = None) -> dict:
        if not config:
            config = content
        return self.__resolve(content, config)

    @abc.abstractmethod
    def __resolve(self, cur_value: Any, config: dict) -> dict:
        raise NotImplementedError



from pathlib import Path
import yaml

VALID_YAML_SUFFIXES = [".yaml", ".yml"]


def load(path: str) -> dict:
    if not any(path.endswith(suffix) for suffix in VALID_YAML_SUFFIXES):
        # Add yaml suffix if the filepath is missing it
        possible_paths = [Path(path + suffix) for suffix in VALID_YAML_SUFFIXES]
        valid_paths = [p for p in possible_paths if p.is_file()]
        if valid_paths:
            valid_path = valid_paths[0]
        else:
            raise FileNotFoundError(f"Unable to resolve {path}")
    else:
        valid_path = Path(path)
        if not valid_path.is_file():
            raise FileNotFoundError(f"Unable to resolve {path}")
    with open(valid_path, 'r') as file:
        content = yaml.safe_load(file)
    return content


def parse_numeric_value(value: str):
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            return value



from __future__ import annotations

import re
from pathlib import Path
from typing import Any, Optional

from yaml_extender import yaml_loader
from yaml_extender.resolver.resolver import Resolver
from yaml_extender.xyml_exception import RecursiveReferenceError, ReferenceNotFoundError, ExtYamlSyntaxError

REFERENCE_REGEX = r'\{\{(.+?)(?::(.*?))?\}\}'
ARRAY_REGEX = r'(.*)?\[(\d*)\]'

MAXIMUM_REFERENCE_DEPTH = 30


class ArithmeticOperation:

    SUPPORTED_FUNCS = {"+": lambda x, y: x + y,
                       "-": lambda x, y: x - y,
                       "*": lambda x, y: x * y,
                       "/": lambda x, y: x / y}
    ARITHMETIC_REGEX = r'(.+)([' + ''.join(["\\" + k for k in SUPPORTED_FUNCS.keys()]) + r'])\s*(\d+)'

    def __init__(self, reference: str, operation: str, value: str):
        self.reference = reference.strip()
        self.operation = operation.strip()
        self.value = yaml_loader.parse_numeric_value(value.strip())

    def __repr__(self):
        return f"{self.reference} {self.operation} {self.value}"

    def apply(self, value):
        num_val = yaml_loader.parse_numeric_value(value)
        return ArithmeticOperation.SUPPORTED_FUNCS[self.operation](self.value, num_val)

    @staticmethod
    def parse(expression: str) -> Optional[ArithmeticOperation]:
        match = re.search(ArithmeticOperation.ARITHMETIC_REGEX, expression)
        if match:
            return ArithmeticOperation(match[1], match[2], match[3])
        else:
            return None


class ReferenceResolver(Resolver):

    def __init__(self, fail_on_resolve: bool = True):
        super().__init__(fail_on_resolve)

    def _Resolver__resolve(self, cur_value: Any, config: dict):
        """Resolves all references in a given value using the provided content dict"""
        new_value = cur_value
        if isinstance(cur_value, dict):
            for k in cur_value.keys():
                cur_value[k] = self._Resolver__resolve(cur_value[k], config)
        elif isinstance(cur_value, list):
            for i, x in enumerate(cur_value):
                cur_value[i] = self._Resolver__resolve(x, config)
        else:
            new_value = self.resolve_reference(cur_value, config)
        return new_value

    def resolve_reference(self, value: Any, config: dict, depth: int = 0) -> Any:
        if not isinstance(value, str) or "{" not in value:
            return value
        if depth > 30:
            raise RecursiveReferenceError(value)
        new_value = value
        # In order to store the full match the whole regex is packed into a group
        for ref_match in re.finditer(REFERENCE_REGEX, value):
            ref = ref_match.group(1).strip()
            default_value = ref_match.group(2)
            if default_value:
                default_value = number_parse(default_value.strip())
            failed = ""
            current_config = config
            # Resolve arithmetic operation
            operation = ArithmeticOperation.parse(ref)
            if operation:
                ref = operation.reference
            # Resolve reference, including subrefs
            for subref in ref.split('.'):
                # Resolve arrays
                arr_match = re.match(ARRAY_REGEX, subref)
                if arr_match:
                    array_name = arr_match[1]
                    try:
                        index = int(arr_match[2])
                    except TypeError:
                        raise TypeError(f"Unable to convert index of '{subref}' to integer.")
                    if array_name in current_config:
                        if len(current_config[array_name]) > index:
                            current_config = current_config[array_name][index]
                        else:
                            failed = array_name + f"[{index}]"
                            break
                    else:
                        failed = array_name
                        break
                else:
                    if subref in current_config:
                        current_config = current_config[subref]
                    else:
                        failed = subref
                        break
            if failed:
                if default_value is not None:
                    ref_val = default_value
                elif self.fail_on_resolve:
                    raise ReferenceNotFoundError(failed)
                else:
                    ref_val = None
            else:
                ref_val = current_config

            if ref_val is not None:
                if operation:
                    ref_val = operation.apply(ref_val)
                if ref_match.group(0) == value:
                    # If the whole string is just a reference return the value without string replacement
                    # in order to preserve float & int types
                    return ref_val
                else:
                    # Replace the reference string with the value
                    new_value = new_value.replace(ref_match.group(0), str(ref_val))

        if new_value == value:
            if self.fail_on_resolve:
                raise ReferenceNotFoundError(value)
            else:
                return value

        # Resolve recursive references
        new_value = self.resolve_reference(new_value, config, depth + 1)
        return new_value


def number_parse(s: str):
    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except ValueError:
            return s



class ExtYamlError(Exception):
    pass


class ReferenceNotFoundError(ExtYamlError):

    def __init__(self, reference):
        self.message = f"Unable to resolve {reference}"


class ExtYamlSyntaxError(ExtYamlError):
    pass


class RecursiveReferenceError(RecursionError):

    def __init__(self, reference):
        self.message = f"Maximum recursive depth reached, while resolving {reference}." \
                       f"Is there a loop in your configuration?"



from __future__ import annotations

import copy
import re
from pathlib import Path
from typing import Any

from yaml_extender.resolver.reference_resolver import ReferenceResolver
from yaml_extender.resolver.resolver import Resolver
from yaml_extender.xyml_exception import RecursiveReferenceError, ReferenceNotFoundError, ExtYamlSyntaxError

LOOP_KEY = "xyml.for"
LOOP_CONTENT_KEY = "xyml.content"
LOOP_REGEX = r'(.+):(.+)'
MAXIMUM_REFERENCE_DEPTH = 30


class LoopResolver(Resolver):

    def __init__(self, fail_on_resolve: bool = True):
        super().__init__(fail_on_resolve)
        self.ref_resolver = ReferenceResolver(False)

    def _Resolver__resolve(self, cur_value: Any, config: dict):
        """Resolves all references in a given value using the provided content dict"""
        new_value = cur_value
        if isinstance(cur_value, dict):
            for k, v in cur_value.items():
                new_value[k] = self._Resolver__resolve(v, config)
            if LOOP_KEY in cur_value:
                new_value = self.resolve_loop(cur_value[LOOP_KEY], new_value, config)
        elif isinstance(cur_value, list):
            for i, x in enumerate(cur_value):
                resolved_loop_content = self._Resolver__resolve(x, config)
                if isinstance(resolved_loop_content, list):
                    # Keep list flat; don't create list of lists
                    del new_value[i]
                    for value in resolved_loop_content:
                        new_value.insert(i, value)
                else:
                    new_value[i] = resolved_loop_content
        return new_value

    def resolve_loop(self, loop_desc, loop_content, config):
        loop_values = []
        # Retrieve value and iterator
        match = re.search(LOOP_REGEX, loop_desc)
        if not match or len(match.groups()) < 2:
            raise ExtYamlSyntaxError(f"No valid loop statement: {loop_desc}")
        iterator = match[1]
        iteration_value = config[match[2]]
        if not isinstance(iteration_value, list):
            raise ExtYamlSyntaxError(f"No valid loop statement: {loop_desc}")

        # Remove loop statement from dict
        del loop_content[LOOP_KEY]
        if LOOP_CONTENT_KEY in loop_content:
            if len(loop_content) > 1:
                raise ExtYamlSyntaxError(f"Flat Loop content does not allow other mapping values: {loop_desc}")
            loop_content = loop_content[LOOP_CONTENT_KEY]
            for item in iteration_value:
                target_value = copy.deepcopy(loop_content)
                target_value = self.ref_resolver.resolve(target_value, {iterator: item})
                loop_values.extend(target_value)
        else:
            for item in iteration_value:
                target_value = copy.deepcopy(loop_content)
                target_value = self.ref_resolver.resolve(target_value, {iterator: item})
                loop_values.append(target_value)
        return loop_values

